import { Part } from '../Nest/Core/Part';
import { Path } from '../Nest/Core/Path';
import { DefaultBin } from '../Nest/Core/DefaultBin';
import { PathGeneratorSingle } from '../Nest/Core/PathGenerator';
import { NestCache } from '../Nest/Core/NestCache';
import { NestDatabase } from '../Nest/Core/NestDatabase';
import { NestFiler } from '../Nest/Common/Filer';
import { OptimizeMachine } from '../Nest/Core/OptimizeMachine';
import { Individual } from '../Nest/Core/Individual';
import Worker from "../Nest/Core/OptimizeWorker.worker";
import { InitClipperCpp } from '../Nest/Common/ClipperCpp';

export class LayoutEngine {
    parts: Part[] = [];
    board: Path;
    threads: Worker[] = [];
    /**
     *
     */
    constructor() {

        //清理缓存,这个缓存是全局的
        PathGeneratorSingle.Clear();
        NestCache.Clear();
        DefaultBin.InsideNFPCache = {};

        //注册bin
        let binPath = DefaultBin;
        // binPath.Id = undefined; //清除这个缓存
        PathGeneratorSingle.RegisterId(binPath);

        this.board = binPath;

    }
    init(parts: Part[]) {
        this.parts = parts;
    }

    async run() {
        await InitClipperCpp();
        let m = new OptimizeMachine;//优化器
        m.Bin = this.board; //指定Bin容器
        m.PutParts(this.parts);//把零件放上去
        let count = 0;
        m.callBack = async (i: Individual) => {
            console.log('优化结果', i);
            count++;
            if(count>10){
                m.Suspend();//关闭优化器
            }
            
        }//当有新结果时,回调

        await m.Start();//启动这个
    }

    async startThread() {

        let db = new NestDatabase();
        db.Paths = PathGeneratorSingle.paths;
        db.Parts = this.parts;
        db.Bin = DefaultBin

        let f = new NestFiler();
        db.WriteFile(f);//写入到文件中
        let best = Infinity;

        let t = new Worker;
        this.threads = [t];

        t.onmessage = (e) => {
            let f = new NestFiler(e.data);
            let inv = new Individual(db.Parts);
            inv.ReadFile(f);
            if (best <= inv.Fitness)
                return;
            best = inv.Fitness;
            let text = `优化率:${inv.Fitness}`;
            console.log(text);
        };
        t.postMessage(f._datas);//Post给Worker

    }

    async stopThread() {
        this.threads[0].terminate()
    }
}